package algds.contest;

/**
 * Speichert die Daten eines Beamers.
 * Ein Beamer ohne Ziel (<code>dest==null</code>) wird halber Beamer
 * genannt. Ein Beamer mit Ziel heisst kompletter Beamer.
 * Halbe Beamer haben alle <code>ID == -1</code>, wird ein Beamer
 * komplettiert, dann wird seine ID auf einen neuen positiven Wert
 * gesetzt.
 * @author Kai-Uwe Sattler, Christian Semrau
 * @see Maze#beamer
 */
public final class Beamer implements Comparable{

/**
 * Konstruiert einen halben Beamer.
 */
public Beamer(int x, int y, int l) {
	X = x;
	Y = y;
	L = l;
	dest = null;
	ID = -1;
	primary = false;
}      

	  

/**
 * Gibt an, ob der Beamer komplett ist, also ob er ein Ziel hat
 * (<code>dest!=null</code>).
 * @return true, wenn der Beamer komplett ist
 */
public boolean isComplete() {
	return dest!=null;
}      

	/** Globaler Zaehler, der die komplettierten Beamer zaehlt.
	 * Wird beim Laden eines Labyrinths zurueckgesetzt (in MazeIO.load).
	 * Wird beim Anlegen eines neuen Labyrinths zurueckgesetzt
	 * (in ControlFrame.mazeNew_actionPerformed).
	 */
	static int counter=0;
	/** Zielbeamer. Ist <code>dest==null</code>, so ist der Beamer
	 * ein halber Beamer. */
	public Beamer dest;
	/** ID des Beamers.
	 * Die ID wird nur fuer die Anzeige benoetigt, fuer die interne
	 * Arbeit ist die ID egal.
	 * Nummern von geloeschten oder getrennten Beamern werden nicht
	 * direkt neu vergeben. Erst wenn ein Labyrinth neu geladen oder
	 * neu erzeugt wird, wird der Zaehler zurueckgesetzt.
	 * Halbe Beamer haben die ID -1.
	 */
	public int ID;
	/** Startpunkt. */
	public int X, Y, L;
	/** Erster Endpunkt des kompletten Beamers.
	 * In einem kompletten Beamer ist genau ein Ende primaer.
	 * In einem halben Beamer ist das Ende nicht primaer. */
	public boolean primary;

/**
 * Konstruiert eine kompletten Beamer und vervollstaendigt
 * den Zielbeamer zu einem kompletten.
 * @param x,y,l Koodinaten des Beamers
 * @param b Zielbeamer
 */
public Beamer(int x, int y, int l, Beamer b) {
	// Startpunkt
	X = x; Y = y; L = l;

	// zur Sicherheit das moeglicherweise vorhandene
	// alte Ziel des Beamers entfernen
	b.removeDest();

	// dieser neue Beamer kriegt die gleiche neue ID wie der
	// bereits vorhandene Zielbeamer
	ID = b.ID = counter++;

	// die beiden Beamer haben sich gegenseitig als Ziel
	this.dest = b;
	b.dest = this;

	setPrimary();
}

/**
 * Entfernt das Ziel eines Beamers.
 * Der Zielbeamer wird ebenfalls von seinem Ziel - diesem Beamer -
 * getrennt.
 * @return der alte Zielbeamer
 */
public Beamer removeDest() {
	Beamer d = dest;
	dest = null;
	ID = -1;
	primary = false;
	if (d!=null){
		// wir hatten ein Ziel, also muessen wir uns vom Ziel trennen
		d.removeDest();
		// Dieser Aufruf fuehrt nicht zu einer Endlosrekursion,
		// weil das Ziel dieses Beamers bereits entfernt ist.
	}
	return d;
}

/**
 * Vergleicht zwei Beamer nach ihren Koordinaten.
 * Kleiner ist der Beamer, der im unteren Level, weiter oben oder
 * weiter links steht (die Kriterien werden in dieser Reihenfolge
 * geprueft).
 * @return int
 * @param o Beamer mit dem verglichen werden soll
 */
public int compareTo(Object o) {
	Beamer b = (Beamer)o;

	int i;
	i = L - b.L;
	if (i!=0) return i;
	i = Y - b.Y;
	if (i!=0) return i;
	i = X - b.X;
	return i;
}

/**
 * Verbindet den Beamer mit dem Zielbeamer zu einem kompletten Beamer.
 * @param d Zielbeamer
 */
public void connect(Beamer d) {

	// zur Sicherheit die moeglicherweise vorhandenen
	// alten Ziele entfernen
	removeDest();
	d.removeDest();

	// dieser Beamer kriegt die gleiche neue ID wie der
	// bereits vorhandene Zielbeamer
	ID = d.ID = counter++;

	// die beiden Beamer haben sich gegenseitig als Ziel
	this.dest = d;
	d.dest = this;

	setPrimary();
}

/**
 * Der "kleinere" Beamer wird auf primary gesetzt.
 * @see #compareTo(java.lang.Object)
 */
protected void setPrimary() {
	if (dest == null) {
		primary = false;
		return;
	}

	boolean p = compareTo(dest)<0;
	primary = p;
	dest.primary = !p;
}

/**
 * 
 * @return java.lang.String
 */
public String toString() {
	return "Beamer["+X+","+Y+","+L+(dest!=null?"-"+dest.X+","+dest.Y+","+dest.L:"")+"]";
}
}