package algds.contest;


/**
 * Basisklasse fuer alle Hamster, die sich im
 * Labyrinth bewegen sollen. Jede Implementierung muss von dieser
 * Klasse abgeleitet werden und die Methode <tt>run</tt>
 * ueberschreiben. Eine Instanz der abgeleiteten Klasse
 * wird in die Labyrinth-Welt geladen und dort gestartet.
 * <br>Alle Konstanten-Werte und Methoden koennen ohne Ankuendigung
 * bis zum 12.03.2001 geaendert werden, der grundlegende Aufbau wird
 * sich aber nicht mehr aendern.
 * <p>
 * <b>Aenderungen:</b>
 * <br>31.01.2001: Erste oeffentliche Version.
 * <br>08.02.2001: Die Konstanten befinden sich nun im Interface
 * algds.contest.Constants, welches von Hamster implementiert wird.
 * Diese Konstanten koennen direkt als <code>CORN</code> angesprochen
 * werden, man muss nicht <code>Constants.CORN</code> schreiben!
 * <br><code>getCorn</code> heisst jetzt <code>getAvailableCorn</code>.
 * <br>05.03.2001: Der Hamster hat stets eine Zeitspanne von
 * {@link Constants#INFINITE_LOOP} Millisekunden, um einen der Befehle
 * {@link #forward()}, {@link #beamMeUp()} oder {@link #turn(int)}
 * aufzurufen. Nach einem solchen Aufruf hat er wieder soviel Zeit usw.
 * Damit soll verhindert werden, dass ein Hamster ewig in einer
 * Endlosschleife haengt, denn beim Aufruf einer der drei Methoden verliert
 * er Energie, so dass er "nach endlicher Zeit" (das kann eine
 * Weile dauern...) verhungert ist.
 * <br>08.03.2001: Neue Hamster-Methoden verbose(), debuglog(String),
 * debuglogln(String). Beim Aufruf von beamMeUp() wird dem Hamster stets
 * Energie abgezogen, auch wenn nicht gebeamt werden kann.
 * <br>30.04.2001: Nach einer bestimmten Gesamtlaufzeit
 * ({@link Constants#TOTAL_RUNTIME}) wird der Lauf des Hamster abgebrochen.
 * Auf diese Weise kann ein Hamster sich nicht 2 Wochen Zeit lassen :-)
 * <br>04.05.2001: Korrektur der Gesamtzeitkontrolle.
 * <p>
 * Wer Fragen hat, stelle die bitte an
 * <a href="mailto:Christian.Semrau@student.uni-magdeburg.de">Christian</a>.
 * <p>
 * Version vom 04.05.2001
 * @author Kai-Uwe Sattler, Christian Semrau, Stephan Finn
 * @see Constants
 */
public class Hamster implements Constants {

	/** Speichert die verwendete MazeEngine, also die Umgebung,
	 * die den Hamster durchs Labyrinth bewegt. */
	private MazeEngine engine = null;
	
	/** Wird geworfen, wenn die Engine null ist, weil der Hamster
	 * gestoppt werden soll. */
	private HamsterStopException hse = new HamsterStopException("STOPPED!");





	/**
	 * Prueft, ob die Engine noch existiert
	 */
	public boolean isEngineAlive() {
		return engine!=null;
	}
    
	/**
	 * Konstruktor zum Erzeugen eines Hamster-Objektes.
	 * Sollte nichts tun!
	 * Alle Initialisierungen und Handlungen des Hamsters
	 * muessen in der Methode {@link #run()} geschehen, denn
	 * die Hamsterbefehle funktionieren nicht, bevor run()
	 * aufgerufen wird.
	 */
	public Hamster() {
	}

	/**
	 * Interne Methode fuer die Ablaufumgebung.
	 * Darf vom Hamster nicht verwendet werden.
	 * <p>
	 * Setzt oder entfernt die Engine des Hamster.
	 * <p>
	 * Sie ist public, um Interessierten die Moeglichkeit zu geben,
	 * eigene MazeEngines zu schreiben und zu verwenden, so wie das
	 * letztes Jahr mit dem MazeRunnerX von Johannes Zander geschah, der
	 * nicht zum Originalpaket gehoerte. (Der sich aber als so gut
	 * erwiesen hat, dass dieses Jahr die ganze Umgebung auf seiner
	 * Engine aufbaut.)
	 * <br>Diese Methode wird waehrend des Wettbewerb nicht mehr
	 * public sein.
	 * @param me die Engine, oder <code>null</code> wenn sie entfernt
	 * werden soll.
	 */
	public void setEngine(MazeEngine me) {
		// nullsetzen hilft hier nichts,
		// weil die Variable direkt danach wieder gesetzt wird
		// engine = null;
		engine = me;
	}

	/**
	 * Nimmt die angegebene Anzahl von Koernern aus dem aktuellen
	 * Feld auf.
	 * Wenn weniger daliegt, nimmt er nur was da ist.
	 * Wenn er weniger ins Maul kriegt, nimmt er nur, was reinpasst.
	 *
	 * @param groesse die Groesse der aufzunehmenden Koerner
	 * @param anzahl die Anzahl der aufzunehmenden Koerner
	 * @return int die Anzahl der tatsaechlich aufgenommenen Koerner
	 * @see Constants#CORNCARRY_MAX
	 * @see Constants#CORNSIZE_MAX
	 */
	public int take(int groesse, int anzahl) {
		if (engine == null) throw hse;
		return engine.take(groesse, anzahl);
	}

	/**
	 * Legt die angegebene Anzahl von Koernern im aktuellen Feld ab.
	 * Wenn er weniger hat, legt er nur ab, was er hat.
	 * Wenn nicht alles aufs Feld passt, wird nur abgelegt, was
	 * draufpasst (alle Felder ausser dem Heimatfeld haben nur Platz
	 * fuer ein Volumen von <code>CORNFIELD_MAX</code>
	 * Groesseneinheiten).
	 *
	 * @param groesse die Groesse der abzulegenden Koerner
	 * @param anzahl die Anzahl der abzulegenden Koerner
	 * @return int die Anzahl der tatsaechlich abgelegten Koerner
	 * @see Constants#CORNFIELD_MAX
	 * @see Constants#CORNSIZE_MAX
	 */
	public int drop(int groesse, int anzahl) {
		if (engine == null) throw hse;
		return engine.drop(groesse, anzahl);
	}

	/**
	 * Frisst die angegebene Anzahl von Koernern.
	 * Diese Koerner muss der Hamster schon im Maul tragen, wenn
	 * er weniger traegt, frisst er alles was er von der Groesse hat.
	 * Die Energie des Hamsters wird dadurch um den Wert
	 * <code>n*groesse*ENERGY_CORN</code>
	 * aufgeladen (wobei <code>n</code> das Minimum aus
	 * <code>anzahl</code> und der Kornzahl [der gegebenen Groesse] ist,
	 * die der Hamster tatsaechlich im Maul traegt),
	 * hoechstens jedoch auf den Wert <code>ENERGY_MAX</code>.
	 *
	 * @param groesse die Groesse der zu fressenden Koerner
	 * @param anzahl die Anzahl der zu fressenden Koerner
	 * @return int die Anzahl der tatsaechlich gefressenen Koerner
	 * @see Constants#ENERGY_CORN
	 * @see Constants#ENERGY_MAX
	 * @see Constants#CORNSIZE_MAX
	 */
	public int eat(int groesse, int anzahl) {
		if (engine == null) throw hse;
		return engine.eat(groesse, anzahl);
	}

	/**
	 * Beamt den Hamster auf ein anderes Feld.
	 * Bringt den Hamster zum "anderen Ende" des Beamers. Das ist
	 * ein Feld, auf dem ein Beamer steht, der den Hamster wieder
	 * zurueckbringt. Wenn kein Beamer auf dem aktuellen Feld ist,
	 * wird auch nicht gebeamt, und <code>false</code> geliefert.
	 * <br>Die Energie aendert sich vor dem Beamen um
	 * <code>ENERGY_STEP</code>, egal ob dann gebeamt wird oder nicht.
	 * <br>Es koennen auch auf Beamerfeldern Maiskoerner liegen
	 * ({@link #look()} und {@link #lookCurrent()} liefern
	 * dann <code>CORN | BEAMER</code> zurueck).
	 * <br>Das Heimatfeld kann ebenfalls einen Beamer haben.
	 *
	 * @return
	 * <code>true</code>, wenn gebeamt werden konnte,
	 * sonst <code>false</code> (kein Beamer da).
	 * @throws HamsterDiedException
	 * wenn der Hamster verhungert ist.
	 * @see Constants#ENERGY_STEP
	 */
	public boolean beamMeUp() throws HamsterDiedException {
		if (engine == null) throw hse;
		return engine.beamMeUp();
	}

	/**
	 * Bewegt den Hamster um einen Schritt in der aktuellen
	 * Richtung vorwaerts. Seine Energie aendert sich nach dem Schritt
	 * um <code>ENERGY_STEP</code> Energiepunkte.
	 * Bei einer Kollision wird der Hamster einmal nach links
	 * gedreht. Seine Energie aendert sich nach der Kollision um
	 * <code>ENERGY_TURN+ENERGY_COLLISION</code> Energiepunkte.
	 *
	 * @return
	 * <code>true</code>, wenn Schritt ausgefuehrt werden konnte,
	 * sonst <code>false</code> (Kollision).
	 * @throws HamsterDiedException
	 * wenn der Hamster verhungert ist.
	 * @see Constants#ENERGY_STEP
	 * @see Constants#ENERGY_COLLISION
	 * @see Constants#ENERGY_TURN
	 */
	public boolean forward() throws HamsterDiedException {
		if (engine == null) throw hse;
		return engine.forward();
	}

	/**
	 * Aendert die Laufrichtung, wobei das aktuelle Feld
	 * nicht verlassen wird. Seine Energie aendert sich nach der Drehung
	 * um <code>ENERGY_TURN</code>.
	 *
	 * @param direction Drehrichtung
	 * (<code>TURN_LEFT</code> oder <code>TURN_RIGHT</code>)
	 * @throws HamsterDiedException wenn der Hamster verhungert ist.
	 * @see Constants#TURN_LEFT
	 * @see Constants#TURN_RIGHT
	 * @see Constants#ENERGY_TURN
	 */
	public void turn(int direction) throws HamsterDiedException {
		if (engine == null) throw hse;
		engine.turn(direction);
	}

	/**
	 * Liefert Informationen ueber das in Blickrichtung
	 * angrenzende Feld.
	 *
	 * @return Eine bitweise-oder-Kombination folgender Werte:<br>
	 * CORN fuer ein Feld mit Koernern<br>
	 * WALL, wenn das Feld in dieser Richtung durch eine Wand begrenzt ist<br>
	 * BEAMER wenn auf dem Feld ein Beamer ist<br>
	 * Liegt keiner der Faelle vor, wird EMPTY geliefert.
	 * @see Constants#EMPTY
	 * @see Constants#CORN
	 * @see Constants#WALL
	 * @see Constants#BEAMER
	 */
	public int look() {
		if (engine == null) throw hse;
		return engine.look();
	}

	/**
	 * Liefert Informationen ueber das Feld, auf dem der Hamster
	 * gerade steht.
	 *
	 * @return Eine bitweise-oder-Kombination folgender Werte:<br>
	 * CORN fuer ein Feld mit Koernern<br>
	 * BEAMER wenn auf dem Feld ein Beamer ist<br>
	 * Liegt keiner der Faelle vor, wird EMPTY geliefert.
	 * @see Constants#EMPTY
	 * @see Constants#CORN
	 * @see Constants#BEAMER
	 */
	public int lookCurrent() {
		if (engine == null) throw hse;
		return engine.lookCurrent();
	}

	/**
	 * Liefert die aktuelle Laufrichtung des Hamsters als
	 * Himmelsrichtung.
	 *
	 * @return Laufrichtung
	 * (<code>DIR_WEST, DIR_NORTH, DIR_EAST, DIR_SOUTH</code>)
	 * @see Constants#DIR_WEST
	 * @see Constants#DIR_NORTH
	 * @see Constants#DIR_EAST
	 * @see Constants#DIR_SOUTH
	 */
	public int getDirection() {
		if (engine == null) throw hse;
		return engine.getDirection();
	}

	/**
	 * Liefert die Anzahl der bisher ausgefuehrten Schritte.
	 *
	 * @return Anzahl der Schritte
	 */
	public int getSteps() {
		if (engine == null) throw hse;
		return engine.getSteps();
	}

	/**
	 * Liefert die Anzahl der bisher gemachten Kollisionen.
	 *
	 * @return Anzahl der Kollisionen
	 */
	public int getCollisions() {
		if (engine == null) throw hse;
		return engine.getCollisions();
	}

	/**
	 * Liefert die aktuelle Energie des Hamsters.
	 * Ein Wert &lt;= 0 bedeutet, dass der Hamster verhungert ist.
	 *
	 * @return Energievorrat des Hamsters
	 */
	public int getEnergy() {
		if (engine == null) throw hse;
		return engine.getEnergy();
	}

	/**
	 * Gibt true zurueck, wenn der Hamster sich auf dem Heimatfeld
	 * befindet, false sonst.
	 * @return Hamster zuhause true/false
	 * @see #getLevel()
	 * @see #getX()
	 * @see #getY()
	 * @see #getHomeLevel()
	 * @see #getHomeX()
	 * @see #getHomeY()
	 */
	public boolean atHome() {
		if (engine == null) throw hse;
		return engine.atHome();
	}

	/**
	 * Liefert die Levelnummer des Heimatfeldes.
	 * Die Ebenen sind aufsteigend nummeriert,
	 * die unterste Ebene hat dabei die Nummer 0.
	 *
	 * @return Levelnummer des Heimatfeldes
	 */
	public int getHomeLevel() {
		if (engine == null) throw hse;
		return engine.getHomeLevel();
	}

	/**
	 * Liefert die X-Koordinate des Heimatfeldes.
	 * Koordinaten koennen positiv oder negativ sein!
	 * @return X-Koordinate des Heimatfeldes
	 */
	public int getHomeX() {
		if (engine == null) throw hse;
		return engine.getHomeX();
	}

	/**
	 * Liefert die Y-Koordinate des Heimatfeldes.
	 * Koordinaten koennen positiv oder negativ sein!
	 * @return Y-Koordinate des Heimatfeldes
	 */
	public int getHomeY() {
		if (engine == null) throw hse;
		return engine.getHomeY();
	}

	/** 
	 * Liefert die Levelnummer der aktuellen Ebene des Labyrinthes. 
	 * Die Ebenen sind aufsteigend nummeriert,
	 * die unterste Ebene hat dabei die Nummer 0.
	 * 
	 * @return Nummer der aktuellen Ebene 
	 */
	public int getLevel() {
		if (engine == null) throw hse;
		return engine.getLevel();
	}

	/**
	 * Liefert die aktuelle x-Position im Labyrinth.
	 * Koordinaten koennen positiv oder negativ sein!
	 *
	 * @return x-Position
	 */
	public int getX() {
		if (engine == null) throw hse;
		return engine.getX();
	}

	/**
	 * Liefert die aktuelle y-Position im Labyrinth.
	 * Koordinaten koennen positiv oder negativ sein!
	 *
	 * @return y-Position
	 */
	public int getY() {
		if (engine == null) throw hse;
		return engine.getY();
	}

	/**
	 * Liefert die Anzahl der Koerner die der Hamster im Maul traegt. <br>
	 * Liefert ein int-Feld der Laenge 1+CORNSIZE_MAX mit folgenden Elementen:<br>
	 * [0] = Gesamtgroesse aller Maiskoerner<br>
	 * [1] = Anzahl der Koerner der Groesse 1<br>
	 * [2] = Anzahl der Koerner der Groesse 2<br>
	 * ...<br>
	 * [CORNSIZE_MAX] = Anzahl der Koerner der Groesse CORNSIZE_MAX<br>
	 *
	 * @return int[] Anzahl der Koerner
	 * @see Constants#CORNSIZE_MAX
	 */
	public int[] getCarriedCorn() {
		if (engine == null) throw hse;
		return engine.getCarriedCorn();
	}

	/** 
	 * Liefert die Anzahl der Koerner gegebener Groesse,
	 * die der Hamster im Maul traegt. <br> 
	 * Parameter groesse -- Ergebnis:<br> 
	 * 0: Gesamtgroesse aller Maiskoerner<br> 
	 * 1: Anzahl der Koerner der Groesse 1<br> 
	 * 2: Anzahl der Koerner der Groesse 2<br> 
	 * ...<br> 
	 * CORNSIZE_MAX: Anzahl der Koerner der Groesse CORNSIZE_MAX<br> 
	 * @param groesse Groesse der Koerner, deren Anzahl geliefert 
	 * werden soll. 
	 * @return int Anzahl der Koerner<br> 
	 * @see Constants#CORNSIZE_MAX
	 */
	public int getCarriedCorn(int groesse) {
		if (engine == null) throw hse;
		return engine.getCarriedCorn(groesse);
	}

	/**
	 * Liefert die Anzahl der gesammelten Koerner auf dem Heimatfeld. <br>
	 * Liefert ein int-Feld der Laenge 1+CORNSIZE_MAX mit folgenden Elementen:<br>
	 * [0] = Gesamtgroesse aller Maiskoerner<br>
	 * [1] = Anzahl der Koerner der Groesse 1<br>
	 * [2] = Anzahl der Koerner der Groesse 2<br>
	 * ...<br>
	 * [CORNSIZE_MAX] = Anzahl der Koerner der Groesse CORNSIZE_MAX<br>
	 *
	 * @return int[] Anzahl der Koerner
	 * @see Constants#CORNSIZE_MAX
	 */
	public int[] getCollectedCorn() {
		if (engine == null) throw hse;
		return engine.getCollectedCorn();
	}

	/** 
	 * Liefert die Anzahl der gesammelten Koerner der gegebenen 
	 * Groesse auf dem Heimatfeld. <br> 
	 * Parameter groesse -- Ergebnis:<br> 
	 * 0: Gesamtgroesse aller Maiskoerner<br> 
	 * 1: Anzahl der Koerner der Groesse 1<br> 
	 * 2: Anzahl der Koerner der Groesse 2<br> 
	 * ...<br> 
	 * CORNSIZE_MAX: Anzahl der Koerner der Groesse CORNSIZE_MAX<br> 
	 * @param groesse Groesse der Koerner, deren Anzahl geliefert 
	 * werden soll. 
	 * @return int Anzahl der Koerner<br> 
	 * @see Constants#CORNSIZE_MAX
	 */
	public int getCollectedCorn(int groesse) {
		if (engine == null) throw hse;
		return engine.getCollectedCorn(groesse);
	}

	/**
	 * Liefert die Anzahl der verfuegbaren Koerner auf dem aktuellen Feld. <br>
	 * Liefert ein int-Feld der Laenge 1+CORNSIZE_MAX mit folgenden Elementen:<br>
	 * [0] = Gesamtgroesse aller Maiskoerner<br>
	 * [1] = Anzahl der Koerner der Groesse 1<br>
	 * [2] = Anzahl der Koerner der Groesse 2<br>
	 * ...<br>
	 * [CORNSIZE_MAX] = Anzahl der Koerner der Groesse CORNSIZE_MAX<br>
	 *
	 * @return int[] Anzahl der Koerner
	 * @see Constants#CORNSIZE_MAX
	 */
	public int[] getAvailableCorn() {
		if (engine == null) throw hse;
		return engine.getAvailableCorn();
	}

	/** 
	 * Liefert die Anzahl der verfuegbaren Koerner der gegebenen 
	 * Groesse auf dem aktuellen Feld. <br> 
	 * Parameter groesse -- Ergebnis:<br> 
	 * 0: Gesamtgroesse aller Maiskoerner<br> 
	 * 1: Anzahl der Koerner der Groesse 1<br> 
	 * 2: Anzahl der Koerner der Groesse 2<br> 
	 * ...<br> 
	 * CORNSIZE_MAX: Anzahl der Koerner der Groesse CORNSIZE_MAX<br> 
	 * @param groesse Groesse der Koerner, deren Anzahl geliefert 
	 * werden soll. 
	 * @return int Anzahl der Koerner<br> 
	 * @see Constants#CORNSIZE_MAX
	 */
	public int getAvailableCorn(int groesse) {
		if (engine == null) throw hse;
		return engine.getAvailableCorn(groesse);
	}

	/**
	 * Liefert die aktuelle Punktzahl des Hamsters.
	 * <br>Punkte =
	 * <br>POINTS_CORN * getCollectedCorn(0)
	 * <br>+POINTS_STEP * getSteps()
	 * <br>+POINTS_COLLISION * getCollisions()
	 * <br>zus&auml;tzlich kommen nach Ende des Laufs hinzu (das wird
	 * der Hamster selbst dann nicht mehr abrufen koennen):
	 * <br>Wenn Absturz oder Zeit&uuml;berschreitung: POINTS_CRASH
	 * <br>sonst Wenn verhungert: POINTS_DIE
	 * <br>sonst Wenn nicht auf Heimatfeld beendet: POINTS_NOTHOME
	 * <br>sonst 0
	 * @return int
	 */
	public int getPoints() {
		if (engine == null) throw hse;
		return engine.getPoints();
	}

	/** 
	 * Gibt eine Meldung in Form eines Strings im Log-Fenster aus. 
	 * Ob tatsaechlich eine Ausgabe erfolgen soll, kann in der "World"
	 * mit dem Menu-Eintrag
	 * <a href="../../worlddoku.html#menuhamster">Hamster/Mute</a>
	 * eingestellt werden.
	 * 
	 * @param msg die auszugegebende Nachricht 
	 */
	public void log(String msg) {
		if (engine == null) throw hse;
		if (!engine.mute())
			engine.log(msg);
	}

	/** 
	 * Gibt eine Meldung in Form eines Strings im Log-Fenster aus.
	 * <br> Fuehrt die folgende Anweisung aus:
	 * <br><code>log(msg+"\n");</code>
	 * 
	 * @param msg die auszugegebende Nachricht
	 * @see #log(java.lang.String)
	 */
	public void logln(String msg) {
		log(msg+"\n");
	}

	/**
	 * Gibt an, ob der Hamster ausfuehrliche Statusmeldungen geben soll.
	 * Der Wert dieser Funktion kann in der "World" mit dem Menu-Eintrag
	 * <a href="../../worlddoku.html#menuhamster">Hamster/Verbose</a>
	 * eingestellt werden.
	 * @return Ausfuehrliche Meldungen geben
	 * @see #debuglog(java.lang.String)
	 */
	public boolean verbose() {
		if (engine == null) throw hse;
		return engine.verbose();
	}
	
	/**
	 * Gibt die Statusmeldung nur dann aus, wenn der Hamster ausfuehrliche
	 * Meldungen geben soll.
	 * <br> Fuehrt die folgende Anweisung aus:
	 * <br><code>if (verbose()) log(msg);</code>
	 * @param msg die auszugebende Meldung
	 * @see #verbose()
	 * @see #log(java.lang.String)
	 */
	public void debuglog(String msg){
		if (verbose()) log(msg);
	}

	/**
	 * Gibt die Statusmeldung nur dann aus, wenn der Hamster ausfuehrliche
	 * Meldungen geben soll.
	 * <br> Fuehrt die folgende Anweisung aus:
	 * <br><code>debuglog(msg+"\n");</code>
	 * @param msg die auszugebende Meldung
	 * @see #debuglog(java.lang.String)
	 */
	public void debuglogln(String msg){
		debuglog(msg+"\n");
	}

	/**
	 * Liefert den Namen des Hamsters.
	 * Kann ueberschrieben werden, um einen anderen Namen als
	 * den Klassennamen zu liefern.
	 */
	public String getName() {
		return getClass().getName();
	}

	/**
	 * Die Hauptmethode des Hamsters.
	 * Durch Ueberschreiben dieser Methode erfolgt
	 * die Bewegung durch das Labyrinth.
	 * <br>
	 * <br>Verhungert der Hamster, dann wird von {@link #forward()},
	 * {@link #beamMeUp()} oder {@link #turn(int)} eine
	 * {@link HamsterDiedException} geworfen.
	 * <br>Soll der Hamster vom Benutzer gestoppt werden,
	 * dann wird eine {@link HamsterStopException} geworfen,
	 * die den Hamster beim naechsten Aufruf einer der
	 * Hamster-Methoden beendet.
	 */
	public void run(){
	}

	/**
	 * Liefert die Anzahl der bisher ausgefuehrten Drehungen.
	 *
	 * @return Anzahl der Drehungen
	 */
	public int getTurns() {
		if (engine == null) throw hse;
		return engine.getTurns();
	}

}