package algds.contest;

/*
Aenderungen:
09.03.2001 ChS: Variablen nach vorn geholt
*/

/**
 * Speichert ein Level des Labyrinths.
 * @author Kai-Uwe Sattler, Christian Semrau
 */
public final class Level implements MazeConstants {

	/** Breite und Hoehe der Ebene. */
	private int width, height;

	/** Die Beamer der Ebene.
	 * Die Dimension bezeichnet beide Richtungen:
	 * <code>beamer[x+y*width]</code>
	 */
	private Beamer[] beamer;

	/** Die Waende der Ebene.
	 * Die Dimension bezeichnet beide Richtungen:
	 * <code>border[x+y*width]</code>
	 */
	private int[] border;

	/** Realkoordinaten des Feldes mit den logischen
	 * Koordinaten (0,0). */
	int centerX, centerY;

	/** Das Getreide der Ebene.
	 * Die erste Dimension bezeichnet beide Richtungen:
	 * <code>corn[x+y*width][groesse]</code>
	 */
	private int[][] corn;

	/** Flags die den Besichtungsgrad der Felder anzeigen.
	 * Die Dimension bezeichnet beide Richtungen:
	 * <code>visfield[x+y*width]</code>
	 */
	private int[] visfield;

/**
 * Initialisiert alle Zellen des Levels.
 * Die Raender werden gesetzt.
 */
protected void initCells() {

	// Felder anlegen
	border = new int[width*height];
	corn = new int[height*width][1+Hamster.CORNSIZE_MAX];
	beamer = new Beamer[width*height];
	visfield = new int[width*height];

	// Zellen initialisieren
	for (int i = 0; i < width; i++) {
		for (int j = 0; j < height; j++) {
			initCell(i,j);
		}
	}

	// Raender setzen
	for (int i = 0; i < width; i++) {
		border[i+0*width] |= BORDER_NORTH;
		border[i+(height-1)*width] |= BORDER_SOUTH;
	}
	for (int j = 0; j < height; j++) {
		border[0+j*width] |= BORDER_WEST;
		border[(width-1)+j*width] |= BORDER_EAST;
	}
}      

public Level(int w, int h) {
	width = w;
	height = h;
	initCells();
}      

public int getWidth() {
	return width;
}      

public int getHeight() {
	return height;
}      

public Cell getCell(int x, int y) {
	if (!posOK(x,y))
		throw new IllegalArgumentException("Level.getCell("+x+","+y+")");
	return new Cell(this,x,y);
}      

/**
 * Entfernt die Waende einer Zelle aus dem Labyrinth.
 * Entfernt auch die "Rueckwaende".
 * Raender werden nicht entfernt!
 * @param x,y Die Koordinaten
 * @param b die Waende (bitweise oder-Kombination von BORDER_XXX)
 * @see MazeConstants#BORDER_WEST
 * @see MazeConstants#BORDER_NORTH
 * @see MazeConstants#BORDER_EAST
 * @see MazeConstants#BORDER_SOUTH
 */
public void removeCellBorder(int x, int y, int b) {
	if (!posOK(x,y))
		throw new IllegalArgumentException("Level.removeCellBorder("+x+","+y+","+b+")");

	// Raender nicht entfernen!
	if (x == 0)
		b &= ~BORDER_WEST;
	if (x == width - 1)
		b &= ~BORDER_EAST;
	if (y == 0)
		b &= ~BORDER_NORTH;
	if (y == height - 1)
		b &= ~BORDER_SOUTH;

	// Waende
	border[x+y*width] &= ~b;

	// Rueckwaende
	if ((b & BORDER_WEST) != 0 && x > 0) {
		border[(x-1)+y*width] &= ~BORDER_EAST;
	}
	if ((b & BORDER_EAST) != 0 && x < width - 1) {
		border[(x+1)+y*width] &= ~BORDER_WEST;
	}
	if ((b & BORDER_NORTH) != 0 && y > 0) {
		border[x+(y-1)*width] &= ~BORDER_SOUTH;
	}
	if ((b & BORDER_SOUTH) != 0 && y < height - 1) {
		border[x+(y+1)*width] &= ~BORDER_NORTH;
	}
}  

/**
 * Setzt die Waende einer Zelle im Labyrinth.
 * Setzt auch die "Rueckwaende".
 * @param x,y Die Koordinaten
 * @param b die Waende (bitweise oder-Kombination von BORDER_XXX)
 * @see MazeConstants#BORDER_WEST
 * @see MazeConstants#BORDER_NORTH
 * @see MazeConstants#BORDER_EAST
 * @see MazeConstants#BORDER_SOUTH
 */
public void addCellBorder(int x, int y, int b) {
	if(!posOK(x,y))
		throw new IllegalArgumentException("Level.addCellBorder("+x+","+y+","+b+")");
	
	// ungueltige Werte ausblenden
	b &= 15;

	// Waende
	border[x+y*width] |= b;

	// Rueckwaende
	if ((b & BORDER_WEST) != 0 && x > 0) {
		border[(x-1)+y*width] |= BORDER_EAST;
	}
	if ((b & BORDER_EAST) != 0 && x < width - 1) {
		border[(x+1)+y*width] |= BORDER_WEST;
	}
	if ((b & BORDER_NORTH) != 0 && y > 0) {
		border[x+(y-1)*width] |= BORDER_SOUTH;
	}
	if ((b & BORDER_SOUTH) != 0 && y < height - 1) {
		border[x+(y+1)*width] |= BORDER_NORTH;
	}
}  

/**
 * Setzt einen Beamer auf das angegebene Feld. Ein eventueller
 * Zielbeamer wird nicht veraendert.
 * @param x,y Koordinaten des Beamers
 * @param b zu setzender Beamer
 */
public void addBeamer(int x, int y, Beamer b) {
	if (!posOK(x,y))
		throw new IllegalArgumentException("Level.addBeamer(" + x + "," + y+")");
	beamer[x+y*width] = b;
}

/**
 * Legt anzahl Koerner der Groesse groesse auf dem Feld x/y ab.
 * @param x int
 * @param y int
 * @param groesse int (1 bis CORNFIELD_MAX)
 * @param anzahl int
 * @throws IllegalArgumentException
 * wenn die Koordinaten nicht passen oder die groesse falsch ist
 */
public void addCorn(int x, int y, int groesse, int anzahl) {
	if (!posOK(x,y) || !sizeSetOK(groesse))
		throw new IllegalArgumentException(
			"Level.addCorn("+x+","+y+","+groesse+","+anzahl+")");
	corn[x+y*width][groesse] += anzahl;
	corn[x+y*width][0] += anzahl*groesse;
}

/**
 * Fuegt den Besuchsstatus hinzu.
 * @param x int
 * @param y int
 * @param state int
 */
public void addVis(int x, int y, int state) {
	if (!posOK(x,y))
		throw new IllegalArgumentException("Level.addVis("+x+","+y+","+state+")");
	visfield[x+y*width] |= state;
}

/**
 * 
 * @return algds.contest.Beamer
 * @param x int
 * @param y int
 */
public Beamer getBeamer(int x, int y) {
	if (!posOK(x,y))
		throw new IllegalArgumentException("Level.getBeamer("+x+","+y+")");
	return beamer[x+y*width];
}

/**
 * 
 * @return int
 * @param x int
 * @param y int
 */
public int getBorders(int x, int y) {
	if (!posOK(x,y))
		throw new IllegalArgumentException("Level.getBorders("+x+","+y+")");
	return border[x+y*width];
}

/**
 * Liefert die Kornzahlen des angegebenen Feldes.
 *
 * @return int[] Ein int-Feld strukturiert wie bei Hamster.getCorn()
 * @see Hamster.getCorn()
 */
public int[] getCorn(int x, int y) {
	if (!posOK(x,y))
		throw new IllegalArgumentException("Level.getCorn("+x+","+y+")");
	return (int[])corn[x+y*width].clone();
}

/**
 * Liefert die Kornzahlen des angegebenen Feldes.
 *
 * @return Anzahl der Koerner gegebener Groesse
 * @see Hamster.getCorn(int)
 */
public int getCorn(int x, int y, int groesse) {
	if (!posOK(x,y) || !sizeGetOK(groesse))
		throw new IllegalArgumentException("Level.getCorn("+x+","+y+","+groesse+")");
	return corn[x+y*width][groesse];
}

/**
 * Initialisiert eine einzelne Zelle.
 * Die Zelle hat keine Koerner, keine Waende, keine Beamer.
 */
protected void initCell(int x, int y) {
	border[x+y*width] = 0;
	for(int i=0; i<=Hamster.CORNSIZE_MAX; i++)
		corn[x+y*width][i] = 0;
	beamer[x+y*width] = null;
}

/**
 * Gibt an, ob eine bestimmte Wand gesetzt ist.
 * @param x
 * @param y
 * @param direction Richtung (WEST bis SOUTH)
 */
public boolean isBorder(int x, int y, int direction) {
	if (!posOK(x,y) || direction < WEST || direction > SOUTH)
		throw new IllegalArgumentException("Level.isBorder("+x+","+y+","+direction+")");
	return (border[x+y*width] & (1 << direction)) != 0;
}

/**
 * Entfernt den Beamer vom angegebenen Feld, der dazugehoerige
 * Zielbeamer wird nicht entfernt.
 * @param x int
 * @param y int
 * @return Beamer der Beamer, der vorher auf dem Feld war
 */
public Beamer removeBeamer(int x, int y) {
	if (!posOK(x,y))
		throw new IllegalArgumentException("Level.removeBeamer("+x+","+y+")");
	Beamer b = beamer[x+y*width];
	beamer[x+y*width] = null;
	return b;
}

/**
 * Entfernt anzahl Koerner der Groesse groesse vom Feld x/y.
 * @param x int
 * @param y int
 * @param groesse int (1 bis CORNSIZE_MAX)
 * @param anzahl int
 * @throws IllegalArgumentException
 * wenn die Koordinaten oder die groesse nicht stimmen
 */
public void removeCorn(int x, int y, int groesse, int anzahl) {
	if (!posOK(x,y) || !sizeSetOK(groesse))
		throw new IllegalArgumentException(
			"Level.removeCorn("+x+","+y+","+groesse+","+anzahl+")");

	// nicht mehr entnehmen als daliegt
	anzahl = Math.min(anzahl, corn[x+y*width][groesse]);
	corn[x+y*width][groesse] -= anzahl;
	corn[x+y*width][0] -= anzahl*groesse;
}

/**
 * Setzt das Feld mit den logischen (0,0)-Koordinaten des Levels.
 * @param x int
 * @param y int
 */
public void setCenter(int x, int y) {
	if (!posOK(x,y))
		throw new IllegalArgumentException("Level.setCenter("+x+","+y+")");

	centerX = x;
	centerY = y;
}

/**
 * Setzt die Kornzahlen auf die uebergebenen Werte.
 *
 * @param anzahlen Ein int-Feld strukturiert wie bei Hamster.getCorn()
 * @see Hamster.getCorn()
 * @throws IllegalArgumentException
 * wenn die Koordinaten oder die Feldlaenge nicht stimmen,
 * oder eine Komponente negativ ist
 */
public void setCorn(int x, int y, int[] anzahlen) {
	if (!posOK(x,y))
		throw new IllegalArgumentException("Level.setCorn("+x+","+y+", int[])");
	if (anzahlen==null || anzahlen.length!=Constants.CORNSIZE_MAX+1){
		String s = anzahlen!=null? ("anzahlen.length="+anzahlen.length):"anzahlen=null";
		throw new IllegalArgumentException("Level.setCorn("+x+","+y+", "+s);
	}

	int ges = 0;
	// Groessen pruefen und aufsummieren
	for (int groesse = 1; groesse <= Constants.CORNSIZE_MAX; groesse++) {
		if (anzahlen[groesse] < 0)
			throw new IllegalArgumentException(
				"Level.setCorn("+x+","+y+", int[])"
				+ ": anzahlen["+groesse+"]="+anzahlen[groesse]);
		ges += anzahlen[groesse] * groesse;
	}
	// die Groessen einzeln setzen
	corn[x+y*width][0] = ges;
	for (int groesse = 1; groesse <= Constants.CORNSIZE_MAX; groesse++) {
		corn[x+y*width][groesse] = anzahlen[groesse];
	}
}

/**
 * 
 * @return int
 * @param x int
 * @param y int
 */
public int getVis(int x, int y) {
	if (!posOK(x,y))
		throw new IllegalArgumentException("Level.getVis("+x+","+y+")");
	return visfield[x+y*width];
}

/**
 * 
 * @return boolean
 * @param x int
 * @param y int
 */
public boolean posOK(int x, int y) {
	return (x>=0 && y>=0 && x<width && y<height);
}

/**
 * 
 * @param state int
 */
public void setVisAll(int state) {
	for(int xy = width*height-1; xy>=0; xy--)
		visfield[xy] = state;
}

/**
 * Gibt an, ob die Korngroesse zum Auslesen erlaubt ist.
 * @return boolean
 * @param groesse Korngroesse
 */
public boolean sizeGetOK(int groesse) {
	return groesse>=0 && groesse<=Constants.CORNSIZE_MAX;
}

/**
 * Gibt an, ob die Korngroesse zum Setzen erlaubt ist.
 * @return boolean
 * @param groesse Korngroesse
 */
public boolean sizeSetOK(int groesse) {
	return groesse>=1 && groesse<=Constants.CORNSIZE_MAX;
}

/**
 * Fuegt einen Rand hinzu oder entfernt Randreihen.
 * Ist ein Parameter positiv, wird leerer Rand hinzugefuegt
 * (die alte Randmauer bleibt erhalten),
 * ist ein Parameter negativ, wird die Anzahl Randreihen entfernt.
 * @param left int
 * @param right int
 * @param top int
 * @param bottom int
 */
public void addBounds(int left, int right, int top, int bottom) {

	// xx2 sind die alten Werte, xx sind die neuen
	
	int width2 = width, height2 = height;
	int centerX2 = centerX, centerY2 = centerY;
	Beamer[] beamer2 = beamer;
	int[] border2 = border;
	int[][] corn2 = corn;
	int[] visfield2 = visfield;

	width = width2+left+right;
	height = height2+top+bottom;

	initCells();

	centerX = centerX2+left; centerY = centerY2+top;
	if (centerX<0) centerX=0;
	if (centerX>=width) centerX=width-1;
	if (centerY<0) centerY=0;
	if (centerY>=height) centerY=height-1;

	for(int x=Math.max(0, 0-left); x<Math.min(width2,width2+right); x++)
	for(int y=Math.max(0, 0-top); y<Math.min(height2,height2+bottom); y++){
		// Uebertragen
		addCellBorder((x+left),(y+top), border2[x+y*width2]);
		corn[(x+left)+(y+top)*width] = corn2[x+y*width2];
		visfield[(x+left)+(y+top)*width] = visfield2[x+y*width2];
	}

}
}