package algds.contest;
import java.util.Vector;

/**
 * Enthaelt die Datenstruktur eines komplettes Labyrinth.
 * @author Kai-Uwe Sattler, Christian Semrau
 */
public final class Maze implements MazeConstants{

	/** Vorgeschlagene Breite fuer ein Labyrinth. */
	public final static int DEFAULT_WIDTH = 20;

	/** Vorgeschlagene Hoehe fuer ein Labyrinth. */
	public final static int DEFAULT_HEIGHT = 20;

	/** Die Levels des Labyrinths. */
	Vector levels;

	/** Liste der kompletten Beamer. */
	Vector beamers;

	/** Liste der halben Beamer. */
	Vector halfBeamers;

	/** Konstrukteur des Labyrinths. */
	String author = "";

	/** Breite der Level. */
	private int width;

	/** Hoehe der Level. */
	private int height;

	/** Reale Heimatkoordinaten. */
	int homex, homey, homeLevel;

	/** Logische Heimatkoordinaten. */
	int logicalHomex, logicalHomey;

	/** Heimatzelle. */
	Cell homeCell;

/**
 * Legt ein neues Labyrinth an.
 * @param w Breite der Level
 * @param h Hoehe der Level
 */
public Maze(int w, int h) {
	super();
	width = w;
	height = h;
	levels = new Vector();
	beamers = new Vector();
	halfBeamers = new Vector();
}  

/**
 * Liefert den i-ten kompletten Beamer aus der Liste der gespeicherten
 * Beamer.
 */
public Beamer getBeamer(int i) {
	try{
		return (Beamer) beamers.get(i);
	}catch(ArrayIndexOutOfBoundsException exc){
		return null;
	}
}          

/**
 * Liefert die Hoehe des Labyrinths.
 */
public int getWidth() {
	return width;
}          

/**
 * Liefert die Hoehe des Labyrinths.
 */
public int getHeight() {
	return height;
}          

/**
 * Liefert das i-te Level.
 * @return Level
 */
public Level getLevel (int i) {
	if (i<0 || i>levels.size())
		throw new IllegalArgumentException("getLevel: Levelindex "
		+i+" nicht in (0 bis "+levels.size()+")");
  	return (Level) levels.elementAt (i);
} 

/**
 * Fuegt ein Level als oberstes hinzu.
 * Alle Level muessen die gleiche Groesse haben!
 * @throws IllegalArgumentException
 * wenn das Labyrinth eine falsche Groesse hat
 */
public void addLevel(Level l) {
	if (l.getWidth()!=width || l.getHeight()!=height)
		throw new IllegalArgumentException("Maze.addLevel: Size level("
		+l.getWidth()+","+l.getHeight()+")"
		+"!=maze("+width+","+height+")");
	levels.add(l);
}                  

/**
 * Setzt die Realkoordinaten des Heimatfeldes.
 * Wenn die Koordinaten ungueltig sind, wird das Heimatfeld
 * auf real-(0,0,0) gesetzt.
 * @param x int
 * @param y int
 * @param level int
 */
public void setHome(int x, int y, int l) {
	if (l>=0 && l<levels.size()
	&& x>=0 && x<width
	&& y>=0 && y<height){
		homex = x;
		homey = y;
		homeLevel = l;
	}else{
		homex = homey = homeLevel = 0;
	}
	refresh();
}

/**
 * Fuegt einen Beamer und seinen eventuellen Zielbeamer ins Labyrinth ein.
 * @param b algds.contest.Beamer
 */
public void addBeamer(Beamer b) {
	Beamer d = b.dest;
	// Beamer ins Labyrinth einfuegen
	getLevel(b.L).addBeamer(b.X,b.Y,b);
	if (d!=null){
		// Zielbeamer auch ins Labyrinth einfuegen
		getLevel(d.L).addBeamer(d.X,d.Y,d);
		// beide Teile als komplette in die Liste einfuegen
		beamers.add(b);
		beamers.add(d);
	}else{
		// den vorhandenen Teil als haben einfuegen
		halfBeamers.add(b);
	}
}

/**
 * Fuegt ein neues leeres Level als oberstes hinzu.
 */
public void addNewLevel() {
	addLevel(new Level(width, height));
}

/**
 * Fuegt den Besuchsstatus hinzu.
 * @param x int
 * @param y int
 * @param z int
 * @param state int
 * @see Level#addVis(int,int,int)
 */
public void addVis(int x, int y, int z, int state) {
	getLevel(z).addVis(x,y,state);
}

/**
 * Liefert ein Cell-Objekt, das auf die angegebenen Koordinaten
 * zugreift.
 */
public Cell getCell(int x, int y, int z) {
	Level l = (Level) levels.elementAt(z);
	return l.getCell(x, y);
}

/**
 * Liefert die Heimatzelle.
 * @return algds.contest.Cell
 */
public Cell getHomeCell() {
	if (homeCell==null) refresh();
	return homeCell;
}

/**
 * Liefert die Levelnummer des Heimatfeldes.
 */
public int getHomeLevel() {
	return homeLevel;
}

/**
 * Liefert die reale X-Koordinate des Heimatfeldes.
 */
public int getHomeX() {
	return homex;
}

/**
 * Liefert die reale X-Koordinate des Heimatfeldes.
 */
public int getHomeY() {
	return homey;
}

/**
 * Liefert die aktuelle Levelzahl.
 * @return int
 */
public int getLevelCount() {
	return levels.size();
}

/**
 * Liefert die logische X-Koordinate des Heimatfeldes.
 * @return int
 */
public int getLogicalHomeX() {
	return logicalHomex;
}

/**
 * Liefert die logische Y-Koordinate des Heimatfeldes.
 * @return int
 */
public int getLogicalHomeY() {
	return logicalHomey;
}

/**
 * Entfernt den halben Beamer an den Koordinaten x/y/z.
 * War der Beamer komplett, so wird das alte Ziel zu einem
 * halben Beamer.
 */
public void removeHalfBeamer(int x, int y, int z) {
	// Beamer holen und aus Labyrinth entfernen
	// der Zielbeamer bleibt hierbei unveraendert
	Beamer b = getLevel(z).removeBeamer(x, y);

	if (b == null)
		return;

	if (b.isComplete()) {
	// kompletter Beamer
		// Zielbeamer ermitteln
		Beamer b2 = b.dest;
		// die beiden Beamer voneinander trennen
		b.removeDest();
		// die beiden nun halben Beamer aus der Liste nehmen
		beamers.remove(b);
		beamers.remove(b2);
		// den nicht zu loeschenden halben behalten
		halfBeamers.add(b2);
	} else {
		// halben Beamer aus Liste entfernen
		halfBeamers.remove(b);
	}
}

/**
 * Entfernt das Level aus dem Labyrinth.
 * Loescht alle Beamer des Levels. Komplette Beamer
 * werden zu halben Beamern. Lag das Heimatfeld in diesem Level,
 * so wird es auf (0,0,0) gesetzt.
 */
public void removeLevel(int z) {
	// das letzte Level nicht entfernen
	if (levels.size()<=1) return;
	Beamer b;

	// alle kompletten Beamer des Level aufspalten und loeschen
	// Beamer aus oberen Levels ein Level runtersetzen
	for (int i = 0; i < beamers.size();) {
		b = (Beamer) beamers.elementAt(i);
		if (z == b.L)
			/**
Beamer aus dem Labyrinth und aus der Liste entfernen und an der
selben Stelle in der Liste weitermachen.
Sollte ein kompletter Beamer mit beiden Enden in der zu loeschenden
Ebene liegen, so finden wir zuerst den, der eher in der Liste
auftritt. Beim Loeschen des Beamers werden beide Enden aus der Liste
entfernt, so dass das andere Ende gar nicht mehr betrachtet wird.
Der Zielbeamer ist dann ein halber und wird im naechsten Schritt
entfernt.
			*/
			removeHalfBeamer(b.X, b.Y, z);
		else{
			if (b.L>z) {
				// Beamer liegt hoeher, also die Levelnummer
				// reduzieren
				b.L--;
			}
			// zum naechsten Element in der Liste gehen
			i++;
		}
	}

	// halbe Beamer des Levels loeschen
	// halbe Beamer aus oberen Levels ein Level runtersetzen
	for (int i = 0; i < halfBeamers.size();) {
		b = (Beamer) halfBeamers.elementAt(i);
		if (z == b.L)
			// Beamer aus dem Labyrinth und aus der Liste entfernen
			// und an der selben Stelle in der Liste weitermachen
			removeHalfBeamer(b.X, b.Y, z);
		else{
			if (b.L>z) {
				// Beamer liegt hoeher, also die Levelnummer
				// reduzieren
				b.L--;
			}
			// zum naechsten Element in der Liste gehen
			i++;
		}
	}

	levels.remove(z);
	if (homeLevel > z)
		// Heimatlevel liegt oberhalb des geloeschten
		homeLevel--;
	else
		if (homeLevel == z) {
			// Heimatfeld lag im geloeschten Level
			homex = homey = homeLevel = 0;
		}
	refresh();
}

/**
 * Liefert den Besuchsstatus.
 * @return int
 * @param x int
 * @param y int
 * @param z int
 * @see Level#getVis(int,int)
 */
public int getVis(int x, int y, int z) {
	return getLevel(z).getVis(x,y);
}

/**
 * Aktualisiert einige Variablen, die zwecks Zeitersparnis
 * zwischengespeichert werden.
 * @see #logicalHomex
 * @see #logicalHomey
 * @see #homeCell
 */
public void refresh() {
	logicalHomex = homex-getLevel(homeLevel).centerX;
	logicalHomey = homey-getLevel(homeLevel).centerY;
	homeCell = getCell(homex, homey, homeLevel);
}

/**
 * Konstruiert den Zielbeamer von b auf dem angegebenen Feld.
 *
 * @param b Startbeamer, der vervollstaendigt werden soll
 * @param x,y,z Koordinaten des Zielbeamers
 * @return der neu erzeugte Zielbeamer
 */
public Beamer addBeamer(Beamer b, int x, int y, int z) {
	if (b.dest!=null) {
		// Beamer hat schon ein Ziel, also dieses Ziel weiterverwenden
		addBeamer(b);
		return b.dest;
	}

	// neuen Zielbeamer konstruieren
	Beamer b2 = new Beamer(x,y,z,b);
	// neuen Beamer ins Level einfuegen
	getLevel(z).addBeamer(x,y,b2);
	// Startbeamer ist kein halber mehr
	halfBeamers.remove(b);
	// zwei komplette anfuegen
	beamers.add(b);
	beamers.add(b2);
	return b2;
}

/**
 * Konstruiert einen halben Beamer auf dem angegebenen Feld.
 * Wenn auf dem Feld schon ein Beamer ist, wird der einfahc geliefert,
 * egal, ob er halb oder komplett ist.
 * @return der halbe Beamer
 */
public Beamer addHalfBeamer(int x, int y, int z) {
	Beamer b = getLevel(z).getBeamer(x,y);
	if (b!=null)
		return b;
	// Beamer anlegen
	b = new Beamer(x, y, z);
	// ins Level tun
	getLevel(z).addBeamer(x,y,b);
	// in die Liste aufnehmen
	halfBeamers.add(b);
	return b;
}

/**
 * Verbindet zwei Beamer.
 * @param b algds.contest.Beamer
 * @param d algds.contest.Beamer
 */
public void connectBeamer(Beamer b, Beamer d) {
	// einer der Beamer fehlt
	if (b==null || d==null) return;

	// die Beamer von ihren Zielen trennen
	disconnectBeamer(b);
	disconnectBeamer(d);

	// miteinander verbinden
	b.connect(d);

	// sind jetzt keine halben mehr, sondern komplette
	halfBeamers.remove(b);
	halfBeamers.remove(d);
	beamers.add(b);
	beamers.add(d);
}

/**
 * Trennt den Beamer von seinem Ziel.
 * @param b algds.contest.Beamer
 * @return der Zielbeamer, oder null, wenn keiner da war
 */
public Beamer disconnectBeamer(Beamer b) {
	// kein Beamer uebergeben oder kein Zielbeamer
	if (b==null || b.dest==null) return null;

	// Ziel merken
	Beamer d = b.dest;

	// ziel trennen
	b.removeDest();

	// sind jetzt kein kompletter Beamer mehr,
	// sondern zwei halbe
	beamers.remove(b);
	halfBeamers.add(b);
	beamers.remove(d);
	halfBeamers.add(d);

	return d;
}

/**
 * Liefert die aktuelle Levelzahl.
 * @return int
 */
public int getBeamerCount() {
	return beamers.size();
}

/**
 * Liefert den i-ten halben Beamer aus der Liste der gespeicherten
 * HalfBeamer.
 * @see EngineXMouse#editBeamer(int,int,boolean)
 */
public Beamer getHalfBeamer(int i) {
	try{
		return (Beamer) halfBeamers.get(i);
	}catch(ArrayIndexOutOfBoundsException exc){
		return null;
	}
}

/**
 * Liefert die aktuelle Levelzahl.
 * @return int
 */
public int getHalfBeamerCount() {
	return halfBeamers.size();
}

/**
 * Gibt an, ob die uebergebenen Koordinaten in der Ebene gueltig sind.
 * @return boolean
 * @param x int
 * @param y int
 */
public boolean posOK(int x, int y) {
	return (x>=0 && y>=0 && x<width && y<height);
}

/**
 * Gbit an, ob die uebergebenen Koordinaten im Labyrinth gueltig sind.
 * @return boolean
 * @param x int
 * @param y int
 */
public boolean posOK(int x, int y, int z) {
	return (x>=0 && y>=0 && x<width && y<height && z>=0 && z<levels.size());
}

/**
 * Entfernt den kompletten oder halben Beamer an den
 * Koordinaten x/y/z mitsamt eventuellem Zielbeamer.
 * @return den gerade geloeschten Beamer
 */
public Beamer removeBeamer(int x, int y, int z) {
	// Beamer holen und entfernen
	// der Zielbeamer wird dabei nicht veraendert
	Beamer b = getLevel(z).removeBeamer(x, y);

	if (b == null)
		return null;

	if (b.isComplete()) {
		// kompletter Beamer
		// auch den Zielbeamer aus dem Labyrinth entfernen
		Beamer d = b.dest;
		getLevel(d.L).removeBeamer(d.X, d.Y);
		beamers.remove(b);
		beamers.remove(d);
	} else {
		// halber Beamer
		halfBeamers.remove(b);
	}
	return b;
}

/**
 * Setzt das komplette Labyrinth auf einen bestimmten Besuchsstatus.
 * @param state int
 */
public void setVisAll(int state) {
	for(int i = 0; i<levels.size(); i++)
		getLevel(i).setVisAll(state);
}

/**
 * Setzt alles auf "gesehen".
 */
public void setVisClear() {
	setVisAll(SAW_ALL);
}

/**
 * Setzt alles auf "ungesehen".
 */
public void setVisHidden() {
	setVisAll(0);
}

/**
 * Fuegt einen Rand hinzu oder entfernt Randreihen.
 * Ist ein Parameter positiv, wird leerer Rand hinzugefuegt,
 * ist ein Parameter negativ, wird die Anzahl Randreihen entfernt.
 * Beamer, die in entferntem Bereich sind, werden entfernt.
 * <br>
 * Ein Aendern der Levelgroesse innerhalb der World ist momentan schwierig,
 * weil die EngineX eine feste Levelgroesse fuer die Anzeige verwendet,
 * und diese nicht direkt umgestellt werden kann (das ist auch der Grund
 * dafuer, dass die Zellengroesse erst beim Neuladen geaendert werden kann).
 * Eine Groessenaenderung ausserhalb der World oder waehrend des Speichern
 * oder Ladens ist dagegen moeglich.
 * <br>
 *
 * @param left int
 * @param right int
 * @param top int
 * @param bottom int
 */
public void addBounds(int left, int right, int top, int bottom) {
	if (width+left+right<=0) return;
	if (height+top+bottom<=0) return;

	for(int i=0; i<levels.size(); i++){
		Level l1 = (Level)levels.elementAt(i);
		l1.addBounds(left, right, top, bottom);
	}

	width += left+right;
	height += top+bottom;

	Vector HB = halfBeamers;
	Vector B = beamers;
	halfBeamers = new Vector();
	beamers = new Vector();

	for(int i=0; i<HB.size(); i++){
		Beamer b = (Beamer)HB.elementAt(i);
		b.X += left; b.Y += top;
		if (posOK(b.X, b.Y))
			addBeamer(b);
	}

	for(int i=0; i<B.size(); i++){
		Beamer b = (Beamer)B.elementAt(i);
		if (b.primary){
			Beamer d = b.dest;
			b.X += left; b.Y += top;
			d.X += left; d.Y += top;
			if (posOK(b.X, b.Y) && posOK(d.X, d.Y))
				addBeamer(b);
		}
	}

	homex += left;
	homey += top;
	if (homex<0) homex=0;
	if (homex>=width) homex=width-1;
	if (homey<0) homey=0;
	if (homey>=height) homey=height-1;
}

/**
 * Vertauscht zwei Levels.
 * @param nr1 int
 * @param nr2 int
 */
public void swapLevel(int nr1, int nr2) {
	// Levels holen, Bereichsfehler werden hier schon erkannt
	Level l1 = getLevel(nr1);
	Level l2 = getLevel(nr2);

	// Levels in der Liste vertauschen
	levels.setElementAt(l1, nr2);
	levels.setElementAt(l2, nr1);

	// Beamer durchgehen
	for (int i = 0; i < beamers.size(); i++) {
		Beamer b = (Beamer) beamers.elementAt(i);
		if (nr1 == b.L)
			b.L = nr2;
		else
		if (nr2 == b.L)
			b.L = nr1;
	}

	// Halfbeamer durchgehen
	for (int i = 0; i < halfBeamers.size(); i++) {
		Beamer b = (Beamer) halfBeamers.elementAt(i);
		if (nr1 == b.L)
			b.L = nr2;
		else
		if (nr2 == b.L)
			b.L = nr1;
	}

	// die Zentren werden in den Levels gespeichert und sind mitgetauscht

	// Heimatfeld anpassen
	if (homeLevel == nr1) homeLevel = nr2;
	else
	if (homeLevel == nr2) homeLevel = nr1;
}
}