package algds.contest;

import javax.swing.JTextArea;

import java.util.Date;

/**
 * Implementiert eine MazeEngine. Bildet die Basis fuer die EngineX,
 * die die grafische Anzeige des Labyrinths realisiert.
 * @author Kai-Uwe Sattler, Christian Semrau, Stephan Finn
 */
public class MazeEngineImpl implements MazeEngine, MazeConstants, Logable {

	/** Das Labyrinth, in dem der Hamster laeuft. */
	public Maze maze;

	/** Der Hamster. */
	Hamster hamster = null;

	/** Semaphore, mit der die aufrufende Methode auf das Ende des
	 * Hamsterlaufes warten kann.
	 * Wird in hamsterStopped() durch notify() freigegeben. */
	Object waiter = null;

	/** Sorgt fuer die Einhaltung der Einzelschritt-Zeiten.. */
	Watch watch = null;

	/** Sorgt fuer die Einhaltung der Gesamtzeit. */
	Watch watchTotal = null;

	/** Gibt nach einem Hamsterlauf an, welche Ausnahme auftrat.
	 */
	String exceptionString = null;
		
	/**
	 * Die Exception, die den Hamster zum Absturz bringt.
	 */
	Exception hamsterException = null;

	/** Aktuelle Schrittzahl. */
	int steps;

	/** Anzahl Kollisionen. */
	int collisions;

	/** Real-Position. Gilt erst, wenn der Hamster geladen ist. */
	int xpos, ypos, zpos;

	/** Richtung als WEST, nicht als DIR_WEST. */
	int direction;

	/** Aktuelle Backenfuellung. */
	int[] carry;

	/** Aktuelle Energie. */
	int energy;

	/** Zusatzpunkte des Hamster fuer
	Verhungern, Nicht zuhause beendet, Absturz oder Timeout. */
	int extraPoints;

	/**
	 * Volumen an Mais, das der Hamster gefressen hat.
	 */
	int ate;

	/** Die TextArea, in der die log-Mitteilungen des Hamsters landen. */
	JTextArea textarea = null;

	/** Startet den Hamster in einem anderen Thread */
	Runner runner = null;

	/** Aktuelles Level, in dem der Hamster ist. */
	Level currLevel;

	/** Wird geworfen, wenn der Hamster verhungert ist. */
	HamsterDiedException hde = new HamsterDiedException("Hamster verhungert.");

	/** Der Hamster laeuft. */
	boolean running = false;

	/** Anzahl gemachter Drehungen. */
	int turns;

	/** Gibt an, ob der Hamster ausfuehrliche Meldungen geben soll.
	 * see Hamster#verbose()
	 */
	boolean verbose = true;

	/** Gibt an, ob der Hamster ueberhaupt Meldungen geben soll.
	 * @see Hamster#log(java.lang.String)
	 */
	boolean mute = false;

public MazeEngineImpl (Maze maze) {
	super();
	this.maze = maze;
	initHamster();
}

/**
 * Der Hamster wird gestoppt, indem ihm die Engine entfernt wird.
 * Beim naechsten Aufruf einer Engine-Methode stoppt er dann.
 * @param kill wenn true, dann wird der Hamster sofort mit
 * {@link Runner#timeOut()} beendet.
 */
public void stopHamster (boolean kill, String desc) {
	if (kill)
	{
		runner.timeOut(desc);
		hamster.setEngine(null);
	}else{
		// beim Beenden durch den Benutzer wird hier die Engine
		// entfernt, und wenn der Hamster dann beendet, dann wird
		// die watch in Runner.run() gestoppt
		// Wenn der Hamster dann trotzdem nicht beendet, wird
		// er irgendwann gekillt

		hamster.setEngine(null);
	}
	// nachdem der Hamster wirklich geendet hat, wird vom Runner
	// die Methode hamsterStopped() aufgerufen
}

/**
 * Wird von Runner aufgerufen, wenn der Hamster gestoppt hat.
 */
public void hamsterStopped() {
	// den waiter aufwecken
	if (waiter!=null)
	synchronized(waiter){
		waiter.notify();
	}
}

public void stopWatch(){
	watch.stop();
	watchTotal.stop();
}

/**
 * Startet den Hamster.
 * @param waiter Ein Objekt, das auf die Beendigung des Hamsters
 * wartet.
 */
public void runHamster (Object waiter) {
	// aktualisiert interne Variablen
	maze.refresh();
	initHamster();

	// Hamster laeuft
	running = true;
	
	// der waiter wird in MazeEngineImpl verwaltet
	this.waiter = waiter;

	runner = new Runner (this);
	runner.setName("Hamster");
	
	watch = new Watch(this, INFINITE_LOOP, "Step");
	watchTotal = new Watch(this, TOTAL_RUNTIME, "Total");

	long time = (new java.util.Date()).getTime();
	watch.startTime = watchTotal.startTime = time;

	watch.start();
	watchTotal.start();
	runner.start();
}

/**
 * Setzt den Hamster auf dem Labyrinth aus.
 */
public void setHamster(Hamster HAMSTER) {
	maze.refresh();

	hamster = HAMSTER;
	if (hamster!=null){
		initHamster();
		hamster.setEngine (this);
	}
}


public int look() {

	// Gesehene Wand aktualisieren
	currLevel.addVis(xpos,ypos, (1 << direction));

	if (currLevel.isBorder(xpos, ypos, direction)) {

		// Da ist ne Wand!
		// Mehr gibts nicht zu sehen.
		return WALL;
	}

	// Nachbarfeld bestimmen
	// wenn da keine Wand war, dann existiert ein Nachbarfeld
	int x = xpos + WALK_LUT[direction][WX];
	int y = ypos + WALK_LUT[direction][WY];

	// Nachbarfeld gesehen
	currLevel.addVis(x,y, SAW_FIELD);
	currLevel.addVis(x,y, (1 << antidir[direction]));

	int result = 0;

	Beamer nb = currLevel.getBeamer(x,y);
	if (nb!=null && nb.isComplete())
		// kompletter Beamer
		result |= BEAMER;

	if (currLevel.getCorn(x,y,0) > 0)
		// Getreide auf dem Feld
		result |= CORN;

	return result;
}

/**
 * Geht einen Schritt vorwaerts.
 * @return true Schritt gemacht, false Kollision
 */
public boolean forward() throws HamsterDiedException {
	if (energy <= 0)
	// Hamster verhungert
		hamsterDied();

	// beamMeUp, forward und turn sind "watchTime-Methoden"
	watch.startTime = (new Date()).getTime();

	if (currLevel.isBorder(xpos, ypos, direction)){
		// Kollision mit Wand
		collisions++;
		turn(TURN_LEFT);
		energy += ENERGY_COLLISION;
		if (energy <= 0)
		// Hamster verhungert
			hamsterDied();
		return false;
	}

	xpos += WALK_LUT[direction][WX];
	ypos += WALK_LUT[direction][WY];
	currLevel.addVis(xpos,ypos, Level.VIS_FIELD);

	// ein Schritt gemacht
	steps++;

	energy += ENERGY_STEP;
	if (energy <= 0)
	// Hamster verhungert
		hamsterDied();

	return true;
}

/**
 * Dreht den Hamster.
 */
public void turn(int dir) throws HamsterDiedException {
	if (energy <= 0)
	// Hamster verhungert
		hamsterDied();

	// beamMeUp, forward und turn sind "watchTime-Methoden"
	watch.startTime = (new Date()).getTime();

	if (dir==TURN_LEFT)
		direction = (direction+3)%4;
	else
	if (dir==TURN_RIGHT)
		direction = (direction+1)%4;
	else
		throw new IllegalArgumentException("turn("+dir+")");
	turns++;

	// Energie fuer die Drehung abziehen
	energy += ENERGY_TURN;
	if (energy <= 0)
	// Hamster verhungert
		hamsterDied();
}

public int getDirection () {
	try{
		return DIR_LUT[direction];
	}catch(ArrayIndexOutOfBoundsException exc){
		throw new HamsterStopException("Error in Engine: internal direction = "+direction);
	}
}

/**
 * Liefert die Levelnummer des Hamsters.
 */
public int getLevel () {
	return zpos;
}

/**
 * Haengt den String ans Textfenster an.
 * Befindet sich der Cursor am Textende, wird er ans neue Textende gesetzt,
 * sonst bleibt er wo er ist.
 */
public void log (String msg){
	if (textarea != null){
		// ist der Cursor am Textende
		boolean atEnd = textarea.getCaretPosition()==textarea.getText().length();
		textarea.append (msg);
		if (atEnd)
			// Cursor wieder ans Textende setzen
			textarea.setCaretPosition(textarea.getText().length());
	}else
		System.out.print(msg);
}

/**
 * Liefert die aktuelle Punktzahl.
 * Liefert den Wert des folgenden Ausdrucks:
 * <pre>
 * getCollectedCorn(0) * POINTS_CORN
 * + steps * POINTS_STEP
 * + collisions * POINTS_COLLISION
 * + extraPoints;
 *
 * Extrapoints enthaelt die zusaetzlichen Punkte:
 * Wenn Absturz oder Zeit&uuml;berschreitung: POINTS_CRASH
 * sonst Wenn verhungert: POINTS_DIE
 * sonst Wenn nicht auf Heimatfeld beendet: POINTS_NOTHOME
 * sonst 0
 * </pre>
 */
public int getPoints () {
	return
		getCollectedCorn(0)*POINTS_CORN
		+steps*POINTS_STEP
		+collisions*POINTS_COLLISION
		+extraPoints;
}

/**
 * Liefert die Levelnummer des Heimatfeldes.
 */
public int getHomeLevel() {
	return maze.homeLevel;
}

/**
 * Liefert die logische X-Koodinate des Heimatfeldes.
 */
public int getHomeX() {
	return maze.logicalHomex;
}

/**
 * Liefert die logische Y-Koodinate des Heimatfeldes.
 */
public int getHomeY() {
	return maze.logicalHomey;
}

/**
 * Liefert die logische X-Koodinate des Hamsters.
 */
public int getX () {
	return xpos - currLevel.centerX;
}

/**
 * Liefert die logische Y-Koodinate des Hamsters.
 */
public int getY () {
	return ypos - currLevel.centerY;
}

/**
 * Gibt true zurueck, wenn der Hamster sich auf dem Heimatfeld
 * befindet, false sonst.
 * @return boolean Hamster zuhause
 * @see Hamster#atHome()
 */
public boolean atHome() {
	// xpos,ypos und homex/homey sind real-Koordinaten
	return xpos == maze.homex
		&& ypos == maze.homey
		&& zpos == maze.homeLevel;
}

/**
 * Aktiviert den Beamer.
 * @see Hamster#beamMeUp()
 */
public boolean beamMeUp() throws HamsterDiedException {
	if (energy <= 0)
	// Hamster verhungert
		hamsterDied();

	// beamMeUp, forward und turn sind "watchTime-Methoden"
	watch.startTime = (new Date()).getTime();

	// Energie abziehen, egal ob gebeamt wird
	energy += ENERGY_STEP;
	if (energy <= 0)
	// Hamster verhungert
		hamsterDied();

	Beamer b = currLevel.getBeamer(xpos,ypos);
	if (b == null || !b.isComplete()){
	// kein kompletter Beamer vorhanden
		return false;
	}

	// erfolgreiches Beamen wird als Schritt gezaehlt
	steps++;

	// Koordinaten aendern
	xpos = b.dest.X;
	ypos = b.dest.Y;
	zpos = b.dest.L;
	currLevel = maze.getLevel(zpos);

	currLevel.addVis(xpos,ypos, Level.VIS_FIELD);

	return true;
}

/**
 * Legt Koerner des Hamsters ab.
 * @param groesse (1 bis CORNSIZE_MAX)
 * @param n
 */
public int drop(int groesse, int n) {

	// kann nur soviel ablegen, wie er hat
	// die Korrektheit des Parameters groesse testen
	try{
		n = Math.min(n, carry[groesse]);
	}catch(ArrayIndexOutOfBoundsException exc){
		throw new IllegalArgumentException("drop("+groesse+","+n+")");
	}catch(IllegalArgumentException exc){
		throw new IllegalArgumentException("drop("+groesse+","+n+")");
	}

	// kann nicht mehr ablegen, als auf das Feld passt
	// Ausnahme: Heimatfeld
	if (!atHome()){
		int spaceLeft = CORNFIELD_MAX-currLevel.getCorn(xpos,ypos,0);
		n = Math.min(n, spaceLeft/groesse);
	}

	// nichts abzulegen
	if (n <= 0)
		return 0;

	// Koerner aus dem Maul des Hamsters entfernen
	carry[groesse] -= n;
	carry[0] -= groesse*n;

	// und auf dem Feld ablegen
	currLevel.addCorn(xpos, ypos, groesse, n);

	return n;
}

/**
 * Frisst Koerner.
 *
 */
public int eat(int groesse, int n) {

	// kann nur soviel fressen, wie er hat
	// die Korrektheit des Parameters groesse testen
	try{
		n = Math.min(n, carry[groesse]);
	}catch(ArrayIndexOutOfBoundsException exc){
		throw new IllegalArgumentException("eat("+groesse+","+n+")");
	}catch(IllegalArgumentException exc){
		throw new IllegalArgumentException("eat("+groesse+","+n+")");
	}

	// nichts zu fressen
	if (n <= 0)
		return 0;

	// Koerner aus dem Maul des Hamsters entfernen
	carry[groesse] -= n;
	carry[0] -= groesse * n;

	// merken, wieviel er gefressen hat
	ate += groesse*n;

	// und Energie auffuellen
	energy += n * groesse * Hamster.ENERGY_CORN;

	// Energie-Maximum einhalten
	energy = Math.min(energy, ENERGY_MAX);

	return n;
}

/**
 * Liefert die Kornzahlen des aktuellen Feldes.
 * @see Hamster.getCorn()
 */
public int[] getAvailableCorn() {
	return currLevel.getCorn(xpos, ypos);
}

/**
 * Liefert die Kornzahlen des aktuellen Feldes.
 * @see Hamster.getCorn(int)
 */
public int getAvailableCorn(int groesse) {
	try{
		return currLevel.getCorn(xpos, ypos, groesse);
	}catch(ArrayIndexOutOfBoundsException exc){
		throw new IllegalArgumentException("getAvailableCorn("+groesse+")");
	}catch(IllegalArgumentException exc){
		throw new IllegalArgumentException("getAvailableCorn("+groesse+")");
	}
}

public int[] getCarriedCorn () {
	return (int[])carry.clone();
}

public int getCarriedCorn(int groesse) {
	try{
		return carry[groesse];
	}catch(ArrayIndexOutOfBoundsException exc){
		throw new IllegalArgumentException("getCarriedCorn("+groesse+")");
	}
}

public int[] getCollectedCorn() {
	return maze.homeCell.getCorn();
}

public int getCollectedCorn(int groesse) {
	try{
		return maze.homeCell.getCorn(groesse);
	}catch(ArrayIndexOutOfBoundsException exc){
		throw new IllegalArgumentException("getCollectedCorn("+groesse+")");
	}catch(IllegalArgumentException exc){
		throw new IllegalArgumentException("getCollectedCorn("+groesse+")");
	}
}

public int getCollisions () {
	return collisions;
}

public int getEnergy() {
	return energy;
}

/**
 * Liefert die reale X-Koordinate des Hamsters.
 */
public int getHamsterX() {
	return xpos;
}

/**
 * Liefert die reale Y-Koordinate des Hamsters.
 */
public int getHamsterY() {
	return ypos;
}

public int getSteps () {
	return steps;
}

/**
 * Wird aufgerufen, wenn der Hamster stirbt.
 * Wirft die HamsterDiedException.
 */
public void hamsterDied() throws HamsterDiedException{
	throw hde;
}

public int lookCurrent () {

	// Feld gesehen
	currLevel.addVis(xpos,ypos, SAW_FIELD);

	int result = 0;

	Beamer nb = currLevel.getBeamer(xpos,ypos);
	if (nb!=null && nb.isComplete())
		// kompletter Beamer
		result |= BEAMER;

	if (currLevel.getCorn(xpos,ypos,0) > 0)
		// Getreide auf dem Feld
		result |= CORN;

	return result;
}

/**
 * 
 * @return int
 */
public int neighX() {
	return xpos + WALK_LUT[direction][WX];
}

/**
 * 
 * @return int
 */
public int neighY() {
	return ypos + WALK_LUT[direction][WY];
}


/**
 * Nimmt Koerner auf.
 * @param groesse (1 bis CORNSIZE_MAX)
 * @param n
 */
public int take(int groesse, int n) {

	// kann nicht mehr nehmen, als daliegt
	// die Korrektheit des Parameters groesse testen
	try{
		n = Math.min(n, currLevel.getCorn(xpos, ypos, groesse));
	}catch(ArrayIndexOutOfBoundsException exc){
		throw new IllegalArgumentException("take("+groesse+","+n+")");
	}catch(IllegalArgumentException exc){
		throw new IllegalArgumentException("take("+groesse+","+n+")");
	}

	// noch frei im Maul
	int spaceleft = CORNCARRY_MAX - carry[0];

	// kann nur soviel nehmen, wie noch reinpasst
	n = Math.min(spaceleft/groesse, n);

	// Koerner dem Hamster geben
	carry[groesse] += n;
	carry[0] += groesse*n;

	// Koerner von Feld entfernen
	currLevel.removeCorn(xpos, ypos, groesse, n);

	return n;
}

/**
 * 
 * @return java.lang.String
 */
public String getDirString() {
	switch(direction){
		case WEST: return "W";
		case NORTH: return "N";
		case EAST: return "E";
		case SOUTH: return "S";
		default: return ""+direction;
	}
}

/**
 * Gibt an, ob die Engine bereits einen Hamster hat.
 * @return true, wenn <code>hamster!=null</code> ist
 */
public boolean hasHamster() {
	return hamster!=null;
}

public int getTurns() {
	return turns;
}

public int getAte() {
	return ate;
}

/**
 * Setzt die Variablen des Hamsters auf Startwerte.
 */
protected void initHamster() {
	steps = 0;
	turns = 0;
	collisions = 0;
	extraPoints = 0;
	carry = new int[1+CORNSIZE_MAX];
	energy = ENERGY_START;
	ate = 0;
	direction = NORTH;
	xpos = maze.homex;
	ypos = maze.homey;
	zpos = maze.homeLevel;
	currLevel = maze.getLevel(zpos);
	exceptionString = null;
	hamsterException = null;
}

public void logln(String s) {
	log(s+"\n");
}

public void setLog (JTextArea t) {
	  textarea = t;
}

/**
 * Gibt an, ob der Hamster ausfuehrliche Meldungen geben soll.
 */
public boolean verbose() {
	return verbose;
}

/**
 * Gibt an, ob der Hamster ueberhaupt Meldungen geben soll.
 */
public boolean mute() {
	return mute;
}

}